/*****************************************************************************
 * FvwmVoice.c - Fvwm module.                                                *
 *               Get voice commands from EARS, and send them to fvwm.        *
 *               By Nir Arzooni & Israel Diamand.                            *
 *****************************************************************************/

#include <stdio.h>
#include <string.h>
#include <signal.h>
#include <fcntl.h>
#include <sys/time.h>

#include "../../fvwm/module.h"
#include "../../version.h"
#include "FvwmVoice.h"

/* global variables */
COMM comm_hash[NUM_OF_ENTRIES];  /* hash table for the commands           */
char PIPE[100];                  /* named pipe between listen & FvwmVoice */
char config[100];                /* configuration file                    */
int fd[2];                       /* pipe between fvwm & module            */
int np_fd;                       /* descriptor for named pipe             */
int listen_pid;                  /* son process pid                       */
char comm[C_LEN];                /* buffer for current command            */
int times;                       /* how much times to execute a command ? */
unsigned long win_id;            /* focus window                          */


/*********************************************
 * the hash function - find entry for a word *
 *********************************************/
int hash (char *word)
{
   if (islower(*word)) return (*word - 'a');
   if (isupper(*word)) return (*word - 'A');
   return FAIL; /* wrong word */
}


/***************************************************************
 * newitem - create a new item and insert it to the hash table *
 ***************************************************************/
int newitem(int times, char* word, char* comm)
{
   COMM* item;
   int entry;

   /* find the entry, allocate memory for the new item */
   if ((entry = hash(word)) == FAIL) return FAIL;
   if ((item = (COMM*)(malloc(sizeof(COMM)))) == NULL) return FAIL;

   /* feel the field of the new item */
   item->times = times;
   strcpy(item->word, word);
   strcpy(item->comm, comm);

   /* insert the item to the hash table. each entry in the hash is 
      a header of a linked list.
   */ 
   item->next = comm_hash[entry].next;
   comm_hash[entry].next = item;
   return SUCCESS;
}


/********************************
 * find - find item in the hash *
 ********************************/
COMM* find(char* word)
{
   int entry, cmp;
   COMM *curr;

   if ((entry = hash(word)) == FAIL) return NULL;     /* find the entry */
   curr = comm_hash[entry].next;                      /* the first item */

   /* search for the item in the list. compare word to the current item */
   while (curr != NULL)
      if (! strcmp(word, curr->word)) return curr;    /* found          */
        else curr = curr->next;
   return NULL;                                       /* not found      */
}


/**********************************
 * print_the_hash - for debugging *
 **********************************/
void print_the_hash()
{
   int l;
   COMM *item;

   for (l = 0; l < NUM_OF_ENTRIES; l++) {
      printf("entry is: %c\n", l + 'a');
      item = comm_hash[l].next;
      while (item != NULL) {
         printf("%s, ",item->word);
         item = item->next;
      }
      printf("\n");
   }
}


/*************************************************************
 * read_config_file - read and parse the configuration file. *
 *                    enter the command to the hash table.   *  
 *************************************************************/
void read_config_file(char *config_file)
{
   FILE* config;
   char word[W_LEN], comm[C_LEN];
   char ch;
   int l, times, w, c, line_no = 1;

   /* initialize the hash table */
   for (l = 0; l < NUM_OF_ENTRIES; l++)
      comm_hash[l].next = NULL;

   config = fopen(config_file, "r"); /* open configuration file for reading */

   do {
      w = c = 0;
      if ((ch = fgetc (config)) == EOF) break;              /* break if EOF */

      /* the first item in the line is the word */
      while (ch == ' ' || ch == '\t') ch = fgetc (config); /* ignore spaces */

      /* the word must begin with A-Z or a-z                                */
      if (!(((ch >= 'A') && (ch <= 'Z')) || ((ch >= 'a') && (ch <= 'z')))) {
         printf("syntax error in config file, line No. %d\n",line_no);
         exit(1);
      }

      do word[w++] = ch;                     /* read until you find a space */
        while ((ch = fgetc (config)) != ' ' && ch != '\t');
      word[w] = 0;                          /* put 0 in the end of a string */

      /* the 2nd item in the line is '1' (ONETIME) or 'f' (FOREVER)         */
      while (ch == ' ' || ch == '\t') ch = fgetc (config); /* ignore spaces */

      if (((times = ch) != FOREVER) && (times != ONETIME)) {
         printf("syntax error in config file, line No. %d\n",line_no);
         exit(1);
      }

      do ch = fgetc (config);
        while (ch == ' ' || ch == '\t');                   /* ignore spaces */

      /* the 3rd item is the command, but it can be miss */
      /* read the command: */
      if (ch != '#') 
         do comm[c++] = ch;
           while ((ch = fgetc (config)) != '\n' && ch != EOF && ch != '#');
      comm[c] = 0;                          /* put 0 in the end of a string */

      /* read the rest of the line - in a case of comment after the command */
      while (ch != EOF && ch != '\n') ch = fgetc (config);

      /* enter the new item to the hash */
      if (newitem(times, word, comm) == FAIL) {
         printf("Failed to enter new item to the hash. abrting ...\n");
         exit;
      }

      line_no++;    /* lines counter */
   } while (ch != EOF);                           /* read until End Of File */

   fclose (config);
}


/******************************************************
 *                 The main function                  *
 ******************************************************/
void main(int argc, char **argv)
{
   char *MyName, *temp, *s;                /* used to save the program name */

   /* Save the program name - used for error messages and option parsing    */
   temp = argv[0];
   if ((s = strrchr(argv[0], '/')) != NULL) temp = s + 1;
   MyName = safemalloc(strlen(temp)+2);
   strcpy(MyName,"*");
   strcat(MyName, temp);

   /* fvwm2 always run its moduls with at least 6 parameters */
   if (argc < 6) {
       fprintf(stderr,"%s Version %s should only be executed by fvwm!\n",
                                                           MyName, VERSION);
       exit(1);
   }

   /* Catch signals */
   signal (SIGPIPE, DeadPipe);                       /* exit when fvwm died */ 
   signal (SIGCHLD, DeadPipe);                     /* exit when listen died */
  
   /* the pipe to fvwm descriptors are arguments 1 & 2 */
   fd[0] = atoi(argv[1]);
   fd[1] = atoi(argv[2]);

   /* ignore all messages from fvwm, except the M_FOCUS_CHANGE message */
   SetMessageMask(fd, ~M_FOCUS_CHANGE);

   /* the pipe and config file are $HOME/.ears/FvwmVoice.[pipe|config] */
   sprintf(PIPE,"%s/.ears/FvwmVoice.pipe",getenv("HOME"));
   sprintf(config, "%s/.ears/FvwmVoice.config",getenv("HOME"));

   read_config_file(config);                      /* read the configuration */
   if ( stat(PIPE) != -1) unlink(PIPE);        /* remove pipe file if exist */
   mknod(PIPE, 010777, 0);                               /* make named pipe */

   /* fork a child process to execute listen. */
   if (! (listen_pid = fork())) {                             /* son's code */
      char *arg[] = {"","-geometry", GEOM,     /* arguments for xterm */
                        "-fn",       FONT,
                        "-bg",       BG_COLOR,
                        "-fg",       FG_COLOR,
                        "-e",        LISTEN,       0};
      execvp(XTERM,arg);
      exit(1);
   }
                                                       /* father's code ... */
   np_fd = open(PIPE, O_RDONLY);                   /* open pipe for reading */
   win_id = 0;                                /* default window is the root */
   times = ONETIME;

   Loop();                                                  /* loop forever */
}


/***********************************************************************
 * Loop - this is infinity loop.                                       *
 *        check if the focus window was changed,                       *
 *        and change the win_id if needed.                             *
 *        read messages from Listen and send a command if needed.      *
 *        then sleep for SLEEP_TIME usec, if times is FOREVER,         *
 *        and again ...                                                *
 ***********************************************************************/
void Loop() 
{
   int first_time = 0;                          /* is this the first time ? */
   unsigned long curr_win;      /* the current window, when command arrives */

   while(1) {

      check_focus_change();
    
      if ((first_time = read_from_listen()) == 1)
          curr_win = win_id;   /* remember the window that the was
                                  the focus window when you get the command */

      /* send the command to fvwm, if the command is new, 
         or the command is with 'FOREVER' type.
      */
      if ((times == FOREVER) || first_time) {
         SendText(fd, comm, curr_win);
         first_time = 0;                          /* now the command is old */
      }

      /* make a delay if the command is 'FOREVER'. this done to enable
         a controled movement of the cursor or a window, etc.
      */
      if (times == FOREVER) 
         usleep(SLEEP_TIME);
   }
}


/***********************************************************************
 * check_focus_change - check if the focus window change.              *
 *            read messages from Fvwm and change the win_id if needed. *
 ***********************************************************************/
void check_focus_change()
{
   unsigned long header[3];                  /* header of message from Fvwm */
   int retval;                               /* return value from select    */
   fd_set in_fdset;                          /* \ these parameters          */
   int fd_width;                             /*  \  are needed for          */
   struct timeval tv;                        /*   \    'select'             */

   /* check if there is a message from Fvwm, and read it.
      we must check before reading, to prevent blocking of 'read'
   */
   fd_width = GetFdWidth();
   FD_ZERO(&in_fdset);
   FD_SET(fd[1],&in_fdset);
   tv.tv_sec = tv.tv_usec = 0;          
                  /* don't wait. return immediately if there's no message */
#ifdef __hpux
   retval = select(fd_width,(int *)&in_fdset, 0, 0, &tv);
#else
   retval = select(fd_width,&in_fdset, 0, 0, &tv);
#endif

   /* if there's a message from fvwm, and the type of the message is
      M_FOCUS_CHANGE (the focus window changed), read it.
   */
   if ((retval > 0) && (FD_ISSET(fd[1], &in_fdset))) {
      unsigned long *body;                         /* body of the message */

      if (ReadFvwmPacket(fd[1],header,&body) > 0) { 

         /* ReadFvwmPacket return the header and the body of the message.
            header[1] is the type of the message.
            body[0] is the new window ID.
            we check that the message is M_FOCUS_CHANGE, altough we masked
            the other types of messages, for more safety.
         */
         if (header[1] == M_FOCUS_CHANGE)
            win_id = body[0];
      }
      free (body);
   }
}


/***********************************************************************
 * read_from_listen - read the pipe from listen, to see if there's     *
 *                    a new comand.                                    *
 *                    return 1 in a case of new command, otherwise 0   *
 ***********************************************************************/
int read_from_listen()
{
   char len;                              /* length of word from named pipe */
   char word[W_LEN];                               /* the current word word */
   COMM* command;                                   /* the command's struct */

   /* first read one character, to see if there's a word, and its length  */
   if ( (read(np_fd, &len, sizeof(char)) == sizeof(char)) &&
        (len >= 'a') && (len < 'a' + W_LEN)) {
           len = len - 'a' + 1;
           read(np_fd, word, len * sizeof(char));        /* read the word */
           word[len] = 0;             /* put NULL to terminate the string */
             
           /* look for the word in the hash. if exist, change the 
              command parameters.
           */
           if (command = find(word)) {
              strcpy(comm, command->comm);
              times = command->times;
              return (1);
           }
   }
   return (0);
}


/***********************************************************************
 * DeadPipe - called when FvwmVoice have to die:                       *
 *            when SIGPIPE (Fvwm died), SIGCHLD (listen died)          *
 *            or when FvwmVoice got wrong command from listen.         *
 ***********************************************************************/
void DeadPipe(int mypid)
{
   fprintf(stderr,"FvwmVoice: dead pipe\n");
   unlink(PIPE);
   kill(listen_pid,SIGKILL);
   kill(mypid,SIGKILL);
   exit(0);
}
